#include <sndfile.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <omp.h> 

#include    "sound.h"
#include    "functions.h"
#include    "basics.h"

#define PI 3.14159265

void generate_degradation(Sound * S,int * burst,int nBursts,int nMax,int p) {
/* burst is a vector of size N containing integer numbers:
 0 : the sample is not missing
 1 : the sample is missing */

int N,i,j,flag,duree,debut;


/* STEP 1 : Generate the locations of the bursts */
N=S->nSamples;
if (nBursts*(nMax+2)+2*p>N)
{
    printf("Impossible to generate the burst: not enough samples\n");
    exit(0);
}

i=0;
j=0;
flag=0; /* Indicate if it is possible to create the current burst */
duree=0; /* Duration of the current burst */
debut=0; /* First sample of the current burst */

for (i=0;i<nBursts;i++) {
        flag=0;
        duree=irand(1,nMax); /* First, set the duration of burst */
        /*duree=nMax;*/
        while (flag==0) {
            debut=irand(p,N-p-1); /* Then, set the first sample of the burst */
            if (debut+duree>N-p) /* If the corresponding burst gets out of boundary -> choose another first sample */
                flag=0;
            else if (sum_ivect(burst,debut-1,debut+duree+1,N)==0) {/* If the burst does not overlap with another burst */
                for (j=debut;j<debut+duree;j++) {
                    burst[j]=1;
                    flag=1;
                }
            }
        }
 }

/* STEP 2 : Generate the degraded samples for the listenable version */
for (i=0;i<N;i++) {
    if (burst[i]==1) {
        S->channel[0][i]+=randn(0,0.5);
    }
}

return;
}


void initialization(Sound * S,int * burst) {

int i;
i=0;
/*#pragma omp parallel for private(i) shared(S,burst)*/
for (i=0;i<S->nSamples;i++) {
    if (burst[i]==1)
        S->channel[0][i]=0;
}

return;
}


void reconstruct_signal(Sound * S,int * burst,int nWindow,int p) {

int N,nHop,nFrames,i,j,k,m;
double  * input_zeropad, 
	* output_zeropad,
	* w,
	* y,
	* a,
	* x;
int 	* burst_zeropad,
	* t;


N=S->nSamples;
nHop=nWindow/4;
nFrames=floor((N+nWindow)/nHop)+1;

i=0;
j=0;
k=0;


/* Zero-padded versions of the audio signal and of the burst*/
input_zeropad=zeros_dvect(nWindow+(nFrames-1)*nHop);
output_zeropad=zeros_dvect(nWindow+(nFrames-1)*nHop);
burst_zeropad=zeros_ivect(nWindow+(nFrames-1)*nHop);
#pragma omp parallel for private(i) shared(input_zeropad,burst_zeropad) 
for (i=0;i<N;i++) {
    input_zeropad[i+nWindow]=S->channel[0][i];
    burst_zeropad[i+nWindow]=burst[i];
}

/* Hamming window of size nWindow */
w=zeros_dvect(nWindow);
/*#pragma omp parallel for private(i) shared(w) */
for (i=0;i<nWindow;i++) {
    w[i]=(0.54-0.46*cos(2*PI*i/nWindow))/(4*0.54);
}


y=NULL;
a=NULL;
t=NULL;
x=NULL;


m=0; /* Number of missing samples on frame i */

#pragma omp parallel for private(y,x,t,a,i,j,k,m) shared(input_zeropad,output_zeropad,burst_zeropad,w)
for (i=0;i<nFrames;i++) {

y=zeros_dvect(nWindow);
a=zeros_dvect(p+1);
t=NULL;
x=NULL;

    for (j=0;j<nWindow;j++) {
        y[j]=input_zeropad[i*nHop+j];
    }

    m=sum_ivect(burst_zeropad,i*nHop+p,i*nHop+nWindow-p,(nFrames-1)*nHop + nWindow);

        if (m>0) {

               x=realloc(x,m*sizeof(double)); /* Indexes of the missing samples on frame i */
               t=realloc(t,m*sizeof(double)); /* Interpolated samples on frame i */

                /* Determination of the indexes of the missing samples on frame i
                 (The first and last p samples are not considered)*/
                k=0;
                for (j=i*nHop+p;j<i*nHop+nWindow-p;j++) {
                    if (burst_zeropad[j]==1) {
                        t[k]=j-i*nHop;
                        k++;
                    }
                }


                update_a(y,p,nWindow,a);
                update_x(y,p,a,t,m,x);

                for (j=0;j<m;j++) {
                    y[t[j]]=x[j];
                }

            }


        for (j=0;j<nWindow;j++) {
            output_zeropad[i*nHop+j]+=y[j]*w[j];
        }

free(x);
free(t);
free(y);
free(a);

}


#pragma omp parallel for private(i) shared(S,output_zeropad)
for (i=0;i<N;i++) {
   S->channel[0][i]=output_zeropad[i+nWindow];
}


free(w);
free(burst_zeropad);
free(input_zeropad);
free(output_zeropad);


}


void update_a(double * y,int p,int nWindow,double * a) {

double *R =zeros_dvect(p+1);
double *a_old=zeros_dvect(p);
double sigma2=0;
double k=0;
int i,j=0;

#pragma omp parallel for private(i) shared(R,y)
for (i=0;i<p+1;i++) {
    R[i]=dot_dvect(y,i,nWindow-1,nWindow,y,0,nWindow-i-1,nWindow)/nWindow;
}


/* Levinson algorithm */

/*Initizalization*/
a_old[0]=-R[1]/R[0];
sigma2=(1-a_old[0]*a_old[0])*R[0];
a[0]=a_old[0];


for (j=1;j<p;j++) {
    /* Calculation of the reflection coeff */
    k=0;
    for (i=0;i<j;i++) {
        k+=a_old[i]*R[j-i];
    }
    k=(R[j+1]+k)/sigma2;

    /*Update*/
    a[j]=-k;
    sigma2=(1-a[j]*a[j])*sigma2;
    for (i=j-1;i>=0;i--) {
        a[i]=a_old[i]+a[j]*a_old[j-i-1];
   }

    for (i=0;i<j+1;i++) {
    a_old[i]=a[i];
    }

}

a[0]=1;
for (i=0;i<p;i++) {
    a[i+1]=a_old[i];
}

free(a_old);
free(R);
return;

}





void update_x(double * y, int p, double * a, int * t, int m, double * x) {

double * b=zeros_dvect(p+1);
double **B=zeros_dmat(m,m);
int i,j=0;

for (i=0;i<p+1;i++) {
    b[i]=dot_dvect(a,0,p-i,p+1,a,i,p,p+1);
}

for (i=0;i<m;i++) {
    for (j=i;j<m;j++) {
            if (abs(t[i]-t[j])<p+1) {
                B[i][j]=b[abs(t[i]-t[j])];
                B[j][i]=B[i][j];
            }
    }
}


 for (i=0;i<m;i++) {
     x[i]=0;
     for (j=-p;j<p+1;j++) {
            if (ispresent(t[i]-j,t,m)==0)
            {
                x[i]+=b[abs(j)]*y[t[i]-j];
            }
     }
 }

for (i=0;i<m;i++) {
    x[i]=-x[i];
}

solve_Cholesky(B,x,m);

free(*B);
free(B);
free(b);

return;

}


void solve_Cholesky(double ** A, double * x, int N) {

/*STEP 1 : Decomposition of A as L'DL where :
- L is a lower triangular matrix
- D is a diagonal matrix
*/

double ** L=zeros_dmat(N,N);
double *  d=zeros_dvect(N);
double *  v=zeros_dvect(N);


int i,j=0;

for (j=0;j<N;j++) {
    if (j>0) {
        for (i=0;i<j;i++) {
            v[i]=L[j][i]*d[i];
        }
        v[j]=A[j][j]-dot_dvect(L[j],0,j-1,N,v,0,j-1,N);
    }
    else {
        v[j]=A[j][j];
    }
    d[j]=v[j];

    if (v[j]==0) {
        printf("Singular Matrix %d",j);
        exit(EXIT_FAILURE);
        return;
    }

    if (j<N-1) {
        for (i=j+1;i<N;i++) {
            L[i][j]=(A[i][j]-dot_dvect(L[i],0,j-1,N,v,0,j-1,N))/v[j];

        }
    }

L[j][j]=1;
}

for (i=0;i<N;i++) {
    for (j=0;j<i+1;j++) {
        L[j][i]=L[i][j];
    }
}



/* STEP 2 : Solving two traingular systems*/
for (i=0;i<N;i++) {
    v[i]=0;
}

for (i=0;i<N;i++) {
    v[i]=(x[i]-dot_dvect(L[i],0,i,N,v,0,i,N));
}

for (i=0;i<N;i++) {
    x[i]=0;
}

for (i=N-1;i>-1;i--) {
    x[i]=(v[i]/d[i]-dot_dvect(L[i],i,N-1,N,x,i,N-1,N));
}


free(*L);
free(L);
free(v);
free(d);

return;


}


